/* This file contains routins to perform DMA operations by the 8237 controller:
 *	i8237_setup:	generalized version of dma_setup()
 *	wait_dma:	poll the controller to wait for TC
 * Only channel 1, 2 and 3 would work on XT's.
 */

#include "kernel.h"
#include "driver.h"

PRIVATE int page_reg[] = {
  0x87, 0x83, 0x81, 0x82, 0x8f, 0x8b, 0x89, 0x8a
};

/* port for low 16 bits of DMA address */
#define dma8_addr(c)		(2 * (c))
#define dma16_addr(c)		(4 * (c) + 0xb0)
/* port for top 4 bits of 20-bit DMA addr */
#define dma8_top(c)		page_reg[c]
#define dma16_top		dma8_top
/* port for DMA count (count =  bytes - 1) */
#define dma8_count(c)		(2 * (c) + 1)
#define dma16_count(c)		(4 * (c) + 0xb2)
/* DMA byte pointer flip-flop */
#define dma_flipflop(c)		((c) < 4 ? 0x0c : 0xd8)
/* DMA mode port and bit patterns */
#define dma_mode(c)		((c) < 4 ? 0x0b : 0xd6)
#define dma_read(c)		((c) & 0x03 | 0x44)
#define dma_write(c)		((c) & 0x03 | 0x48)
/* DMA init port and bit patterns */
#define dma_init(c)		((c) < 4 ? 0x0a : 0xd4)
#define dma_reset_val(c)	((c) | 0x04)	/* mask */
#define dma_set_val(c)		((c) & 0x03)	/* clear */
/* DMA status port TC bits */
#define dma_status(c)		((c) < 4 ? 0x08 : 0xd0)
#define dma_tc(c)		(1 << ((c) & 0x03))

/*===========================================================================*
 *				i8237_setup				     *
 *===========================================================================*/
PUBLIC void i8237_setup(ch, op, dma, count)
int ch, op;
phys_bytes dma, count;
{
/* The IBM PC can perform DMA operations by using the DMA chip.  To use it,
 * the DMA (Direct Memory Access) chip is loaded with the 20-bit memory address
 * to be read from or written to, the byte count minus 1, and a read or write
 * opcode.  This routine sets up the DMA chip.  Note that the chip is not
 * capable of doing a DMA across a 64K boundary (e.g., you can't read a
 * 512-byte block starting at physical address 65520).
 */

  /* Set up the DMA registers.  (The comment on the reset is a bit strong,
   * it probably only resets the floppy channel.)
   */
  out_byte(dma_init(ch), dma_reset_val(ch));	/* reset the dma controller */
  out_byte(dma_flipflop(ch), 0);		/* write anything to reset it */
  out_byte(dma_mode(ch), op == DEV_SCATTER ? dma_write(ch) : dma_read(ch));
  if (ch < 4) {
	out_byte(dma8_addr(ch), (int)((dma >> 0) & 0xff));
	out_byte(dma8_addr(ch), (int)((dma >> 8) & 0xff));
	out_byte(dma8_top(ch), (int)((dma >> 16) & 0xff));
	out_byte(dma8_count(ch), (int)(((count - 1) >> 0) & 0xff));
	out_byte(dma8_count(ch), (int)(((count - 1) >> 8) & 0xff));
  } else { /* Oops!  Are these correct? */
	out_byte(dma16_addr(ch), (int)((dma >> 1) & 0xff));
	out_byte(dma16_addr(ch), (int)((dma >> 9) & 0xff));
	out_byte(dma16_top(ch), (int)((dma >> 16) & 0xfe));
	out_byte(dma16_count(ch), (int)(((count - 2) >> 1) & 0xff));
	out_byte(dma16_count(ch), (int)(((count - 2) >> 9) & 0xff));
  }
  out_byte(dma_init(ch), dma_set_val(ch));	/* some sort of enable */
}


/*===========================================================================*
 *				wait_dma				     *
 *===========================================================================*/
PUBLIC void wait_dma(ch)
int ch;
{
  /* poll and wait until TC is set */

  while(!(in_byte(dma_status(ch)) & dma_tc(ch)))
	;
}

